
xVersion" 2012.10.10.1"

module Flash

#document Flash
  // Flash device interface
  //
  // This module uses the spi1 primitives for interrupt driven
  // data exchange over the spi1 bus.
  //
  // The bus is used to interface :
  //
  //   Spansion S25FL004A
  //   4 megabit (512 KByte) CMOS 3.0 Volt Flash Memory
  //   with 50 MHz SPI bus
  //   page size   256 Bytes - for PP
  //   Sector size  65 kByte - for SE
  //   Either mode SPI 0 or mode 3
  //
  // Requiered processor SPI setup ( see [[hwconfig]] )
  //
  //  SPI mode 0 selected, PIC configuration :
  //
  //    SSPSTAT1.SMP  = 0
  //    SSPSTAT1.CKE  = 1
  //    SSPCON1.CKP   = 0
#endDoc


private

  uses
    spi1
    tmr1
    Fatal
    Utils
  endUses


  // //////////////////////////////////////////////////////////////////////////
  //
  // Hardware coupling :

  alias spi1.rx.clear         Flash1RxClear   (   --    ) // Flush receive buffer
  alias spi1.rx.hasData       Flash1RxHasData (   -- b  ) // Get amount of RX bytes available
  alias spi1.rx.read          Flash1RxRead    (   -- b  ) // Read a received byte
  alias spi1.tx.hasRoom       Flash1TxHasRoom (   -- b  ) // Query room in TX buffer
  alias spi1.tx.write         Flash1TxWrite   ( b --    ) // Write a byte to be transmitted
  alias spi1.tx.clear         Flash1TxClear   (   --    ) // Flush transmit buffer
  alias spi1.rx.expectCount!  Flash1RxExpect! ( b --    ) // Set the number of RX bytes to expect
  alias spi1.rx.expectCount@  Flash1RxExpect@ (   -- b  ) // Get the number of RX bytes to expect
  alias spi1.tx.expectCount!  Flash1TxExpect! ( b --    ) // Set the number of TX bytes to expect
  alias spi1.tx.expectCount@  Flash1TxExpect@ (   -- b  ) // Get the number of TX bytes to expect
  alias spi1.start            Flash1Start     (   --    ) // Start the SPI machine
  alias spi1.tx.done?         Flash1TxDone?   (   -- f  ) // Check to see if all transmission done


  // //////////////////////////////////////////////////////////////////////////
  //
  // Device opcodes :

    // ### constant flopName                // Address Bytes - Dummies - data count - meaning
    //
    // For data count (can be both Rx or Tx ) :
    //
    //   n1 = 1 .. infinite
    //   n2 = 1 .. 3
    //   n3 = 1 .. 256
    //

    // Read

      $03 constant  flopRead                // 3 - 0 - n1 - Read data bytes
      $0b constant  flopFastRead            // 3 - 1 - n1 - Read data bytes at higher speed
      $9f constant  flopRDID                // 0 - 0 - n2 - Read identification

    // Write control

      $06 constant  flopWREN                // 0 - 0 -  0 - Write enable
      $04 constant  flopWRDI                // 0 - 0 -  0 - Write disable

    // Erase

      $d8 constant  flopSE                  // 3 - 0 -  0 - Sector erase
      $c7 constant  flopBE                  // 0 - 0 -  0 - Bulk (chip) erase

    // Program

      $02 constant  flopPP                  // 3 - 0 - n3 - Page program

    // Status register

      $05 constant  flopRDSR                // 0 - 0 - n1 - Read from status register
      $01 constant  flopWRSR                // 0 - 0 -  1 - Write to status register

    // Power saving

      $b9 constant  flopDP                  // 0 - 0 - 0  - Deep power down
      $ab constant  flopRES                 // 0 - 3 - n1 - Release from deep power down
                                            // (optionally read electronic signature)



  // //////////////////////////////////////////////////////////////////////////
  //
  // Status bits

    %00000001 constant  flstWIP             // Write in Progress
                                            //   1 - Device busy, a Write Status
                                            //       Register, program or erase
                                            //       operation is in progress.
                                            //   0 - Ready, device is in standby mode
                                            //       and can accept commands.

    %00000010 constant  flstWEL             // Write Enable Latch
                                            //   1 - Device accepts Write Status
                                            //       Register, program or erase
                                            //       commands.
                                            //   0 - Ignores Write Status Register,
                                            //       program and erase commands.

    %00000100 constant  flstBP0             // Block protect
    %00001000 constant  flstBP1             // Block protect
    %00010000 constant  flstBP2             // Block protect
                                            //   000 - protects (0) none
                                            //   001 - protects (1) SA7
                                            //   010 - protects (2) SA7 .. SA6
                                            //   011 - protects (4) SA7 .. SA4
                                            //   100 - protects (8) SA7 .. SA0
                                            //   101 - protects (8) SA7 .. SA0
                                            //   110 - protects (8) SA7 .. SA0
                                            //   111 - protects (8) SA7 .. SA0


    %10000000 constant  flstSRWD            // Status Register Write Disable
                                            //   1 - Protects when w# pin is low.
                                            //   0 - No protection, even when w# pin
                                            //       is low.



  // //////////////////////////////////////////////////////////////////////////
  //

  // values for FlashxProcessState
  enum
    flIdle                                  // Process is doing nothing
    flErasing                               // Waiting for flash erase to complete
    flWriting                               // Waiting for flash write to complete
    flWritingPaged                          // Waiting for first flash page write to complete
    flReadingRam                            // Waiting for flash read to RAM to complete
    flChecking                              // Waiting for flash check to complete
  endEnum


  variable      Flash1ProcessState          // State of process driver
  variable      Flash1Counter               // A counter used in programming
  variable      Flash1Remaining             // Remaining count for paged write
 2variable      Flash1DeviceAddr            // Device address for paged write
  variable      Flash1HighByte              // <> 0 when high byte access is needed

  varInterface  Flash1ProcessState
  varInterface  Flash1Counter
  varInterface  Flash1Remaining
 dvarInterface  Flash1DeviceAddr
  varInterface  Flash1HighByte

  bytePtr       Flash1Ram                   // Flash1Ram and Flash1RamPtr


  // //////////////////////////////////////////////////////////////////////////
  //
  // Variable interfacing

  : Flash1ProcessGoIdle         ( -- ) flIdle         Flash1ProcessState! ;
  : Flash1ProcessGoErasing      ( -- ) flErasing      Flash1ProcessState! ;
  : Flash1ProcessGoWriting      ( -- ) flWriting      Flash1ProcessState! ;
  : Flash1ProcessGoWritingPaged ( -- ) flWritingPaged Flash1ProcessState! ;
  : Flash1ProcessGoReadingRam   ( -- ) flReadingRam   Flash1ProcessState! ;
  : Flash1ProcessGoChecking     ( -- ) flChecking     Flash1ProcessState! ;


  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  : Flash1Ramc@++<> ( b -- f )

    Flash1Ramc@++ <>
  ;


  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  : Flash1Counter--0 ( -- )

    #// If counter > 0 decrement

    Flash1Counter@ 0>
    if
      Flash1Counter--
    endif

  ;


  // //////////////////////////////////////////////////////////////////////////
  //
  // Deferred functions (staticly bound callbacks)

  defer Flash1OperationDone (  --  )        // Callback for completed operation


  // //////////////////////////////////////////////////////////////////////////
  //
  // Low level support words

  : Flash1RxReadDone? ( -- f )

    #// True when all read

    Flash1RxExpect@ 0=
  ;

  // //////////////////////////////////////////////////////////////////////////

  : Flash1SetErrorVerify ( -- )

    flerVerify1 raise
  ;

  // //////////////////////////////////////////////////////////////////////////

  : Flash1SendAddress ( dFlashAddr bOpCode -- )

    $ff and                                 // Mask out byte
    Flash1TxWrite                           // Transmit opcode

    $ff and                                 // Keep low byte
    Flash1TxWrite                           // Addr high byte
    dup >< $ff and                          // isolate high byte
    Flash1TxWrite                           // Addr mid  byte
    $ff and                                 // keep low byte
    Flash1TxWrite                           // Addr low  byte
  ;

  // //////////////////////////////////////////////////////////////////////////

  : ?Flash1ReadError ( -- )

    #// Raise exception on read timeout

    Flash1ReadTimer.fired
    flerReadTimeout1 ?Abort
  ;

  // //////////////////////////////////////////////////////////////////////////

  : ?Flash1WriteError ( -- )

    #// Raise exception on write timeout

    Flash1WriteTimer.fired
    flerWriteTimeout1 ?Abort
  ;

  // //////////////////////////////////////////////////////////////////////////

  : ?Flash1Timeout ( -- )

    #// Raise exception on timeout

    Flash1Timeout.fired
    flerFlashTimeout1 ?Abort
  ;

  // //////////////////////////////////////////////////////////////////////////

  : Flash1ReadStatus ( -- b )

    flopRDSR Flash1TxWrite
    1 Flash1TxExpect!
    1 Flash1RxExpect!
    Flash1Start                             // Start flash machine
    5. Flash1ReadTimer.start                // set timeout
    begin
      ?Flash1ReadError                      // Raise exception on read timeout
      Flash1RxReadDone?
    until
    Flash1ReadTimer.stop
    Flash1RxRead
  ;

  // //////////////////////////////////////////////////////////////////////////

  : Flash1ReadId ( -- )

    #// Debug code

    flopRDID Flash1TxWrite
    1 Flash1TxExpect!
    3 Flash1RxExpect!
    Flash1Start                             // Start flash machine
    5. Flash1ReadTimer.start                // set timeout
    begin
      ?Flash1ReadError                      // Raise exception on read timeout
      Flash1RxReadDone?
    until
    Flash1ReadTimer.stop
    Flash1RxRead drop                       // Manufacturer ID
    Flash1RxRead drop                       // Device       ID
    Flash1RxRead drop                       // Memory capacity
  ;

  // //////////////////////////////////////////////////////////////////////////

  : Flash1WriteDone? ( -- f )

    #// Check if write operation finished

    Flash1ReadStatus
    flstWIP and 0=
  ;

  // //////////////////////////////////////////////////////////////////////////

  : Flash1WriteStatus ( b -- )

    #// Write the device status register

    flopWRSR Flash1TxWrite
             Flash1TxWrite
    2 Flash1TxExpect!
    0 Flash1RxExpect!
    Flash1Start                             // Start flash machine
    50. Flash1WriteTimer.start              // start timeout
    begin
      ?Flash1WriteError                     // Raise exception on write timeout
      Flash1TxDone?
      if
        Flash1WriteDone?
      else
        False
      endIf
    until
  ;

  // //////////////////////////////////////////////////////////////////////////

  : Flash1EnableWrite ( -- )

    flopWREN Flash1TxWrite
    1 Flash1TxExpect!
    0 Flash1RxExpect!
    Flash1Start                             // Start flash machine
    50. Flash1WriteTimer.start              // start timeout
    begin
      ?Flash1WriteError                     // Raise exception on write timeout
      Flash1TxDone?
    until
  ;

  // //////////////////////////////////////////////////////////////////////////

  : Flash1OperationFlagDone ( -- )

    Flash1ProcessGoIdle                     // Reset state to idle
    Flash1OperationDone                     // Inform forground about completion
  ;

  // //////////////////////////////////////////////////////////////////////////

  : Flash1FillTxBuffer ( -- )

    #// Fill Tx buffer as much as possible

    Flash1Counter@ 0<>
    if
      begin
        Flash1Counter@ 0<>                  // More data to send out ?
        Flash1TxHasRoom                     // Buffer space available ?
        and
      while
        Flash1Ramc@++                       // Get data byte
        Flash1TxWrite                       //   write it to the buffer
        Flash1Counter--0                    //   and decrement to_do count
      repeat
      Flash1Start                           // And restart the flash machine
    endIf
  ;

  // //////////////////////////////////////////////////////////////////////////
  //
  // High level support words

  : Flash1Erase ( -- )

    Flash1EnableWrite                       // Turn on write enable
    Flash1ProcessGoErasing                  // Change state
    flopBE Flash1TxWrite
    0 Flash1RxExpect!                       // Set expect count to zero
    1 Flash1TxExpect!
    Flash1Start                             // Start flash machine
    50.000 Flash1Timeout.start              // Start a 50 s timeout
  ;

  // //////////////////////////////////////////////////////////////////////////

  defer DoFlashProcess1 ( -- )


  // //////////////////////////////////////////////////////////////////////////

  : Flash1ReadBlockRam ( dFlashAddr wCount wRamAddr -- )

    #// Start a block read to RAM

    Flash1ProcessGoReadingRam               // Change state
    Flash1RamPtr!                           // Store RAM address
    Flash1HighByte.clear                    // Start with low byte
    >r                                      // Save bCount
    flopRead                                // Opcode
    Flash1SendAddress                       // Transmit opcode and dFlashAddr
    r> Flash1RxExpect!                      // Set expect count to bCount
    4 Flash1TxExpect!
    Flash1Start                             // Start flash machine
    5000 Flash1Timeout.start                // Start a timeout

    begin                                   // Drive reader until finished
      Flash1ProcessState@
      flIdle <>
    while
      DoFlashProcess1
    repeat
  ;


  // //////////////////////////////////////////////////////////////////////////

  : Flash1WriteBlock ( wRamAddr dFlashAddr wCount -- )

    #// Start a block write

    Flash1EnableWrite                       // Turn on write enable
    Flash1ProcessGoWriting                  // Change state
    Flash1Counter!                          // Setup counter
    d>r                                     // Save dFlashAddr
    Flash1RamPtr!                           // Setup RAM pointer
    Flash1HighByte.clear                    // Start with low byte
    dr>                                     // Get dFlashAddrr back
    flopPP                                  // Opcode
    Flash1SendAddress                       // Transmit opcode and dFlashAddr
    0 Flash1RxExpect!
    Flash1Counter@ 4+
    Flash1TxExpect!
    Flash1FillTxBuffer                      // Fill Tx buffer as much as possible
                                            //   and start flash machine
    2000 Flash1Timeout.start                // Start a timeout
  ;

  // //////////////////////////////////////////////////////////////////////////

  : Flash1WriteBlockPaged ( wRamAddr dFlashAddr wCount -- )

    #// Start a block write

    Flash1Counter!                          // Keep a copy of bCount, desired count
    0 4 Pick -                              // Subtract lowest device address byte from zero
                                            //   which is the max amount to write in the first page
    dup Flash1Counter@ >=                   // Check if possible count >= desired count
    over 0= or                              // or if possible count = 0, meaning 256
    if                                      // Yes, can do unpaged write
      Drop                                  // Drop possible count
      Flash1Counter@                        // Get desired count back on stack
      Flash1WriteBlock                      // Perform an unpaged write
    else                                    // No, trickery needed - possible count is on stack still
      Flash1Counter@ over -                 // Get remaining count (= desieed count - possible count)
                                            //   over possible count
      Flash1Remaining!                      // Sstore as remaining count
      Flash1Counter!                        // Store possible count as well
      dDup                                  // Dup dFlashAddrr
      Flash1Counter@ 0                      // Get possible count as double
      d+                                    //   add to copy of dFlashAddr
      Flash1DeviceAddr!                     //   and store as device address
                                            // Start writing the first page
                                            //   analogue to FlashWriteBlock
      Flash1EnableWrite                     // Turn on write enable
      Flash1ProcessGoWritingPaged           // Change state
      d>r                                   // Save dFlashAddr
      Flash1HighByte.clear                  // Start with low byte
      Flash1RamPtr!                         // Setup RAM pointer
      dr>                                   // Get dFlashAddrr back
      flopPP                                // Opcode
      Flash1SendAddress                     // Transmit opcode and dFlashAddr
      Flash1Counter@ 4+
      Flash1TxExpect!
      Flash1FillTxBuffer                    // Fill Tx buffer as much as possible
                                            //   and start the flash machine
      2000 Flash1Timeout.start              // Start a 2 s timeout
    endif
  ;

  // //////////////////////////////////////////////////////////////////////////

  : Flash1CheckBlock ( wRamAddr dFlashAddr wCount -- )

    #// Start a block verification

    Flash1ProcessGoChecking                 // Change state
    >r d>r                                  // Save bCount and dFlashAddr
    Flash1HighByte.clear                    // Start with low byte
    Flash1RamPtr!                           // Store wRamAddr for later use
    dr>                                     // Get dFlashAddr back
    flopFastRead                            // Opcode
    Flash1SendAddress                       // Transmit opcode and dFlashAddr
    0 Flash1TxWrite                         // Write a dummy byte
    r> Flash1RxExpect!                      // Set expect count to bCount
    5 Flash1TxExpect!
    Flash1Start                             // Start flash machine
    5000 Flash1Timeout.start                // Start a timeout
  ;

  // //////////////////////////////////////////////////////////////////////////

  : Flash1AbortRx ( -- )

    #// Abort read operation

    0 Flash1RxExpect!
    Flash1RxClear
  ;

  // //////////////////////////////////////////////////////////////////////////

  : InitFlash1 ( -- )

    #// Module initialization

    Flash1TxClear
    Flash1AbortRx
    Flash1ProcessGoIdle
  ;


  // //////////////////////////////////////////////////////////////////////////
  //
  // Process handlers

  : Flash1ProcessIdle ( -- )

    #// Process is doing nothing

    // Nothing
  ;

  // //////////////////////////////////////////////////////////////////////////

  : Flash1ProcessErasing ( -- )

    #// Waiting for flash erase to complete

    ?Flash1Timeout                          // Reboot on timeout
    Flash1TxDone?                           // Transmit buffer empty ?
    if
      Flash1WriteDone?
      if
        Flash1OperationFlagDone             // Go idle and flag completion to foreground
        Flash1Timeout.stop
      endIf
    endif
  ;

  // //////////////////////////////////////////////////////////////////////////

  : Flash1ProcessWriting ( -- )

    #// Waiting for flash write to complete

    ?Flash1Timeout                          // Reboot on timeout
    Flash1Counter0=                         // More bytes to buffer ?
    if
      Flash1TxDone?                         // Transmit buffer empty ?
      if
        Flash1WriteDone?                    // Device ready ?
        if
          Flash1OperationFlagDone           // Go idle and flag completion to foreground
          Flash1Timeout.stop
        endIf
      endif
    else
      Flash1TxDone?
      if
        Flash1FillTxBuffer                  // Fill Tx buffer as much as possible
      endIf
    endif
  ;

  // //////////////////////////////////////////////////////////////////////////

  : Flash1ProcessWritingPaged ( -- )

    #// Waiting for flash write to complete

    ?Flash1Timeout                          // Reboot on timeout
    Flash1Counter0=                         // More bytes to buffer ?
    if
      Flash1TxDone?                         // Transmit buffer empty ?
      if
        Flash1WriteDone?                    // Device ready ?
        if                                  // 1st page done, start 2nd
          Flash1HighByte.clear              // Start with low byte
          Flash1RamPtr@
          Flash1DeviceAddr@
          Flash1Remaining@
          Flash1WriteBlock
        endIf
      endif
    else
      Flash1TxDone?
      if
        Flash1FillTxBuffer                  // Fill Tx buffer as much as possible
      endIf
    endif
  ;

  // //////////////////////////////////////////////////////////////////////////

  : Flash1ProcessReadingRam ( -- )

    #// Waiting for flash read to RAM to complete

    ?Flash1Timeout                          // Reboot on timeout
    Flash1RxHasData                         // Check for available read data
    if
      begin
        Flash1RxHasData                     // Read all data available
      while
        Flash1RxRead                        // Read data
        Flash1Ramc!++                       // Store it and increment pointer
      repeat
    else
      Flash1RxReadDone?                     // Check if read operation completed
      if                                    // Yes, flag it
        Flash1OperationFlagDone             // Go idle and flag completion to foreground
        Flash1Timeout.stop
      endIf
    endif
  ;

  // //////////////////////////////////////////////////////////////////////////

  : Flash1ProcessChecking ( -- )

    #// Waiting for flash checking to complete

    ?Flash1Timeout                          // Reboot on timeout
    Flash1RxHasData 0>                      // Check for available read data
    if
      Flash1RxRead                          // Read data
      Flash1Ramc@++<>                       // Compare against ram data and advance RAM pointer
      if                                    // Error
        Flash1AbortRx                       // Abort read operation
        Flash1SetErrorVerify                // Flag a verification error condition
        Flash1OperationFlagDone             // Go idle and flag completion to foreground
      endif
    else
      Flash1RxReadDone?                     // Check if read operation completed
      if                                    // Yes, flag it
        Flash1OperationFlagDone             // Go idle and flag completion to foreground
        Flash1Timeout.stop
      endIf
    endif
  ;

  // //////////////////////////////////////////////////////////////////////////

  : Flash1Process ( -- )

    Flash1ProcessState@
    case

      flIdle of                             // Process is doing nothing
          Flash1ProcessIdle
        endOf

      flErasing of                          // Waiting for flash erase to complete
          Flash1ProcessErasing
        endOf

      flWriting of                          // Waiting for flash write to complete
          Flash1ProcessWriting
        endOf

      flWritingPaged of                     // Waiting for paged flash write to complete
          Flash1ProcessWritingPaged
        endOf

      flReadingRam of                       // Waiting for flash read to RAM to complete
          Flash1ProcessReadingRam
        endOf

      flChecking of                         // Waiting for flash check to complete
          Flash1ProcessChecking
        endOf

    default
      drop                                  // Invalid state
      Flash1ProcessGoIdle                   // reset state
    endCase
  ;
    implements DoFlashProcess1


// ////////////////////////////////////////////////////////////////////////////

public

  export  InitFlash1            (                             --    ) // Init module, or reset it

  export  Flash1Erase           (                             --    ) // Start a complete flash erase
  export  Flash1ReadBlockRam    ( dFlashAddr wCount wRamAddr  --    ) // Start a block read to RAM
  export  Flash1WriteBlock      ( wRamAddr dFlashAddr wCount  --    ) // Start a block write
  export  Flash1WriteBlockPaged ( wRamAddr dFlashAddr wCount  --    ) // Start a auto paged block write
  export  Flash1CheckBlock      ( wRamAddr dFlashAddr wCount  --    ) // Start a block verification
  export  Flash1Process         (                             --    ) // hmm ...

  // Deferred words (staticly bound callbacks)

  export  Flash1OperationDone   (                             -- b  ) // Callback for completed operation, passes error status

endModule

